// FILE         : fifo_async_bind.sv
// LICENSE      : Personal/academic use only. Commercial use requires a Commercial License.
// DATE         : 2026-06-16
// PURPOSE      : Binds fifo_async_sva to fifo_async for FPV.
//-----------------------------------------------------------------------------
`default_nettype none

bind fifo_async fifo_async_sva #(
    .WIDTH_G  (WIDTH_G),
    .DEPTHEXP (DEPTHEXP)
) u_sva (
    .wclk              (wclk),
    .wrst_n            (wrst_n),
    .wen_n             (wen_n),
    .wdata             (wdata),
    .wfull             (wfull),
    .rclk              (rclk),
    .rrst_n            (rrst_n),
    .ren_n             (ren_n),
    .rdata             (rdata),
    .rempty            (rempty),
    .waddr_wrside      (waddr_wrside),
    .raddr_rdside      (raddr_rdside),
    .waddr_rdside      (waddr_rdside),
    .raddr_wrside      (raddr_wrside),
    .wrbincount_wrside (wrbincount_wrside),
    .rdbincount_rdside (rdbincount_rdside),
    .rdbincount_wrside (rdbincount_wrside)
);

`default_nettype wire